"""Perform data conversions."""

from pathlib import Path


def path_to_string(path_name: Path, levels: int = 4) -> str:
    """Convert the path to an elided version of the path as a string."""
    # break the path into parts
    parts = path_name.parts
    # note that this uses the - 1 after len
    # because the first part is always the root
    # (for instance, on Unix-like systems, '/'
    # and on Windows, 'C:\')
    if len(parts) - 1 > levels:
        start_index = len(parts) - levels
        return Path("<...>", *parts[start_index:]).as_posix()
    else:
        return path_name.as_posix()



"""Enumerations for defining levels or options for the execexam tool."""

from enum import Enum


class AdviceMethod(str, Enum):
    """An enumeration of the ways in which advice can be sought from an LLM."""

    api_key = "apikey"
    api_server = "apiserver"


class Theme(str, Enum):
    """An enumeration of the themes for syntax highlighting in rich."""

    ansi_dark = "ansi_dark"
    ansi_light = "ansi_light"


class ReportType(str, Enum):
    """An enumeration of report types furnishing details about the exam outcomes."""

    all = "all"
    exitcode = "status"
    debug = "debug"
    finalresult = "result"
    setup = "setup"
    testcodes = "code"
    testfailures = "failure"
    testtrace = "trace"
    testadvice = "advice"



"""Display results from running the execexam tool."""

from typing import Any, Dict, List, Optional

from rich.console import Console
from rich.panel import Panel
from rich.syntax import Syntax

from . import enumerations


def make_colon_separated_string(arguments: Dict[str, Any]):
    """Make a colon separated string from a dictionary."""
    return "\n" + "".join(
        [f"- {key}: {value}\n" for key, value in arguments.items()]
    )


def get_display_return_code(return_code: int, fancy: bool) -> str:
    """Determine the return code from running the specified checks(s)."""
    message = "\n"
    # no errors were found in the executable examination
    if return_code == 0:
        message += "[green]\u2714 All checks passed."
    # there was an error in the executable examination
    else:
        message += "[red]\u2718 One or more checks failed."
    if fancy:
        message += "\n"
    return message


def display_tldr(console: Console) -> None:
    """Display a list of example commands and their descriptions."""
    console.print(
        "[bold yellow]Too Lazy; Didn't Read: Example Commands[/bold yellow]\n"
    )
    console.print(
        "[bold red]Please ensure you are in the directory with the pyproject.toml file to run these commands.[/bold red]\n"
    )

    console.print(
        "[bold cyan]poetry run execexam <path-to-project> <path-to-tests>[/bold cyan]"
    )
    console.print(
        "    Run executable exam for a project with the specified test files."
    )

    console.print(
        "[bold cyan]poetry run execexam <path-to-project> <path-to-tests> --mark <mark>[/bold cyan]"
    )
    console.print("    Run the tests with the specified mark(s).")

    console.print(
        "[bold cyan]poetry run execexam <path-to-project> <path-to-tests> --maxfail[/bold cyan]"
    )
    console.print("    Limit the number of test failures before stopping.")

    console.print(
        "[bold cyan]poetry run execexam <path-to-project> <path-to-tests> --report <report_type>/<all>[/bold cyan]"
    )
    console.print(
        "    Generate the specified type(s) of reports after the exam. Use 'all' to generate all available report types."
    )

    console.print(
        "[bold cyan]poetry run execexam <path-to-project> <path-to-tests> --advice-model <model> --advice-method <method>[/bold cyan]"
    )
    console.print(
        "    Use specified LLM model and method for providing advice on test failures."
    )

    console.print(
        "[bold cyan]poetry run execexam <path-to-project> <path-to-tests> <--debug>/<--no-debug>[/bold cyan]"
    )
    console.print("    Display or disable debugging information.")

    console.print(
        "[bold cyan]poetry run execexam <path-to-project> <path-to-tests> <--fancy>/<--no-fancy>[/bold cyan]"
    )
    console.print("    Display or disable fancy output formatting.")

    console.print(
        "\n[bold yellow]help:[/bold yellow] Use [bold yellow]--help[/bold yellow] to see more options."
    )


def display_advice(return_code: int) -> str:
    """Determine the return code from running the specified checks(s)."""
    message = "\n"
    # no errors were found in the executable examination
    # and no advice is needed, so display helpful message
    if return_code == 0:
        message += "[green]\u2714 Advise requested, but none is needed!"
    # there was an error in the executable examination;
    # note that this is not going to be normally displayed
    else:
        message += "[red]\u2718 Advise requested, and will be provided!"
    # add an extra newline to ensure suitable spacing;
    # note that this is always done and thus there is
    # no need to know whether or not fancy output needed
    message += "\n"
    return message


def display_content(  # noqa: PLR0913
    console: Console,
    display_report_type: enumerations.ReportType,
    report_types: Optional[List[enumerations.ReportType]],
    content: str,
    label: str,
    richtext: bool,
    syntax: bool,
    syntax_theme: str = "ansi_dark",
    syntax_language: str = "python",
    newline: bool = False,
) -> None:
    """Display a diagnostic message using rich or plain text."""
    if report_types is not None and (
        display_report_type in report_types
        or enumerations.ReportType.all in report_types
    ):
        # rich text was chosen and thus the message
        # should appear in a panel with a title
        if richtext:
            # add an extra newline in the output
            # to separate this block for a prior one;
            # only needed when using rich text
            if newline:
                console.print()
            # use rich to print highlighted
            # source code in a formatted box
            if syntax:
                source_code_syntax = Syntax(
                    "\n" + content,
                    syntax_language,
                    theme=syntax_theme,
                )
                console.print(
                    Panel(
                        source_code_syntax,
                        expand=False,
                        title=label,
                    )
                )
            # use rich to print sylized text since
            # the content is not source code
            # that should be syntax highlighted
            else:
                console.print(
                    Panel(
                        content,
                        expand=False,
                        title=label,
                        highlight=True,
                    )
                )
        # plain text was chosen but the content is
        # source code and thus syntax highlighting
        # is needed, even without the panel box
        elif not richtext and syntax:
            source_code_syntax = Syntax(
                "\n" + content,
                syntax_language,
                theme=syntax_theme,
            )
            console.print(f"{label}")
            console.print(source_code_syntax)
        # plain text was chosen and the content is
        # not source code and thus no syntax highlighting
        # is needed and there is no panel box either
        else:
            console.print(f"{label}\n{content}")



"""This module contains the pytest plugin for the execexam package."""

from typing import Any, List, Tuple

import pytest
from _pytest.config import Config
from _pytest.nodes import Item

# create the report list of
# dictionaries that are organized by nodeid
reports: List[dict[str, Any]] = []

# No longer used but may be needed {{{

# internal_coverage = coverage.Coverage()

# }}}


def pytest_configure(config: Config):
    """Define the order marker that can control test order in the test suites."""
    # note that if the plugin did not define the order
    # marker then this would lead to warnings when execexam
    # runs a provided test suite that uses this marker;
    # defining the marker inside of the plugin avoids needing
    # to define it in every test suite that uses this plugin
    config.addinivalue_line(
        "markers", "order(number): Mark test to run in a specific order"
    )


def extract_single_line(text: str) -> str:
    """Extract a single line from a string."""
    # split the text into lines
    lines = text.splitlines()
    # extract the first line from the text
    # and add an ellipsis to indicate that
    # the text has been truncated if there
    # were other lines beyod the first one
    output = lines[0]
    if len(lines) > 1:
        output += " ..."
    return output


def extract_exception_details(call: pytest.CallInfo) -> Tuple[int, str, str]:
    """Process an exception into relevant details about the exact issue and a message."""
    # initialize all of the variables
    # to empty values before processing
    lineno = 0
    exact = ""
    message = ""
    # dealing with a pytest.CallInfo that is an exception and it can
    # first be processed generally and then, if possible, as an AssertionError
    if call.excinfo is not None and isinstance(call.excinfo.value, Exception):
        # extract the line number
        last_traceback_entry = call.excinfo.traceback[-1]
        lineno = last_traceback_entry.lineno + 1
        # transform the exception into a string
        exception_output = str(call.excinfo.value)
        # specifically dealing with an AssertionError and thus
        # there is specialized information that should be extracted
        if isinstance(call.excinfo.value, AssertionError):
            # extract the error message before the "assert" keyword
            message = exception_output.split("assert")[0].strip()
            # there is no message and thus we must
            # set it to a default value of "AssertionError"
            if message == "":
                message = type(call.excinfo.value).__name__
            # extract any details after the "assert" keyword that
            # corresponds to the exact assertion that failed
            exact = exception_output.split("assert")[1].strip()
        # dealing with an exception that is not an AssertionError
        else:
            # note that there is not an ideal match for the information
            # that is available to the plugin for a general purpose
            # exception and the information that the plugin can report
            # --> the message is the exception's text-based rendering
            message = exception_output
            # --> the exact message is the type of the exception
            exact = type(call.excinfo.value).__name__
    # return the details extracted from the exception
    return (lineno, exact, message)


def pytest_collection_modifyitems(items: List[Item]):
    """Reorder the tests based on the 'order' mark that has an integer value."""
    # sort the items in-place based on the 'order' mark;
    # note that this actually modifies the item list which
    # is what pytest uses to determine the order of tests
    items.sort(
        key=lambda item: item.get_closest_marker("order").args[0]  # type: ignore
        if item.get_closest_marker("order")
        else float("inf")
    )


def pytest_runtest_call(item: Item):
    """Called before the test function is called."""
    # save the name of the test
    test_name = item.nodeid
    # save the full path of the test file
    test_file_name_path = item.fspath
    _ = (test_name, test_file_name_path)
    # start the coverage collection
    # sys.settrace(trace_calls)


def pytest_runtest_protocol(item: Item, nextitem: Item):  # type: ignore
    """Track when a test case is run."""
    global reports  # noqa: PLW0602
    # reference the nextitem parameter
    # that is not used by the hook
    _ = nextitem
    # create a new dictionary for the report
    # and add it to the report list
    reports.append({"nodeid": item.nodeid})


def pytest_exception_interact(node: Item, call: pytest.CallInfo, report: Any):
    """Interacts with exceptions."""
    global reports  # noqa: PLW0602
    # reference the report parameter
    # that is not used by the hook
    _ = report
    # extract the details about the exception that was thrown
    exception_info = call.excinfo
    # set the details about the exception to be the empty string
    # and if there are more details about it, then extract them
    traceback_text = ""
    if exception_info is not None:
        traceback_text = exception_info.exconly()
    # there was an assertion error and thus
    # the plugin must extract details about what failed
    if isinstance(call.excinfo.value, Exception):  # type: ignore
        # create an empty dictionary for the test report
        current_test_report = {}
        # find the test report for this specific test that
        # has a passing assertion
        for current_report in reports:
            # found the test report for this specific test
            # based on what matches according to the nodeid
            if current_report["nodeid"] == node.nodeid:
                current_test_report = current_report
        # extract the details about the exception
        (lineno, expl, orig) = extract_exception_details(call)
        # one of the test reports was found
        # and thus we can store information about this assertion
        if current_test_report != {}:
            # create an empty dictionary for the data about
            # the assertions for this failing test
            current_assertion_dict = {}
            # indicate that the assertion failed
            current_assertion_dict["Status"] = "Failed"
            # there is no data about assertions for this test
            if current_test_report.get("assertions") is None:
                # add the needed fields about the assertion
                current_assertion_dict["Line"] = str(lineno)
                current_assertion_dict["Exact"] = extract_single_line(expl)
                current_assertion_dict["Message"] = orig
                # create a new list and add the dictionary with
                # the details about this assertion to the new list
                assertions_dictionary_list = [current_assertion_dict]
                current_test_report["assertions"] = assertions_dictionary_list
            # there is already data about assertions for this test
            else:
                # add the needed fields about the assertion
                current_assertion_dict["Line"] = str(lineno)
                current_assertion_dict["Exact"] = extract_single_line(expl)
                current_assertion_dict["Message"] = orig
                # there is an existing list of assertion dictionaries
                # for this test case and thus we must add a new dictionary
                # to the list that already exists with assertion information
                current_test_report["assertions"].append(
                    current_assertion_dict
                )
        # there was no information about this exception; this would normally
        # occur when there is an underlying problem with running this specific
        # test because otherwise a different hook would have already added
        # the name of the test into the report list. This means that we need
        # to record all of the information about this test in a new report
        else:
            # create a new dictionary for the failing test case
            new_failing_test_report = {}
            new_failing_test_report["nodeid"] = node.nodeid
            # create an empty dictionary for the data about
            # the assertions for this failing test
            current_assertion_dict = {}
            current_assertion_dict["Status"] = "Failed"
            current_assertion_dict["Message"] = traceback_text
            # store the details about this test failure's assertions
            # inside of the assertion dictionary; note that this is
            # essentially "overloading" the assertion dictionary because
            # there are actually no assertions being recorded --- it is
            # only the fact that the test failed and then the traceback
            # of the exception that was raised when running the test
            new_failing_test_report["assertions"] = [current_assertion_dict]  # type: ignore
            # add the new failing test report to the list of reports
            reports.append(new_failing_test_report)


def pytest_assertion_pass(
    item: Any, lineno: int, orig: str, expl: str
) -> None:
    """Extract and save information about a passing assertion."""
    # Important note: this function is only active in the
    # pytest plugin if the project that is using execexam
    # enables the pytest assertion_pass_hook;
    # reference: https://docs.pytest.org/en/stable/reference/reference.html
    global reports  # noqa: PLW0602
    # create an empty dictionary for the test report
    current_test_report = {}
    # find the test report for this specific test that
    # has a passing assertion
    for current_report in reports:
        # found the test report for this specific test
        # based on what matches according to the nodeid
        if current_report["nodeid"] == item.nodeid:
            current_test_report = current_report
    # one of the test reports was found
    # and thus we can store information about this assertion
    if current_test_report != {}:
        # create a dictionary to store details
        # about the passing assertion for this test
        current_assertion_dict = {}
        # indicate that the assertion passed
        current_assertion_dict["Status"] = "Passed"
        # there is no data about assertions for this test
        if current_test_report.get("assertions") is None:
            # create an empty dictionary for the data about
            current_assertion_dict["Line"] = str(lineno)
            current_assertion_dict["Code"] = orig
            current_assertion_dict["Exact"] = extract_single_line(expl)
            # create a new list and add the dictionary with
            # the details about this assertion to the new list
            assertions_dictionary_list = [current_assertion_dict]
            current_test_report["assertions"] = assertions_dictionary_list
        # there is already data about assertions for this test
        else:
            # create an empty dictionary for the data about
            # this assertion and then add the needed fields
            current_assertion_dict["Line"] = str(lineno)
            current_assertion_dict["Code"] = orig
            current_assertion_dict["Exact"] = extract_single_line(expl)
            # there is an existing list of assertion dictionaries
            # for this test case and thus we must add a new one to it
            current_test_report["assertions"].append(current_assertion_dict)


# No longer used but may be needed {{{

# def trace_calls(frame: FrameType, event: str, arg: Any):
#     """Trace function calls."""
#     if event != "call":
#         return
#     code = frame.f_code
#     func_name = code.co_name
#     func_filename = code.co_filename
#     if func_name == "write":
#         # ignore write() calls from print statements
#         return
#     # note that all of this is current hard-coded
#     target_dir = Path(
#         "/home/gkapfham/working/teaching/github-classroom/algorithmology/executable-examinations/solutions/algorithm-analysis-final-examination-solution/exam/questions/"
#     )
#     if Path(func_filename).resolve().parent == target_dir.resolve():
#         called = (func_name, func_filename)
#         _ = called

# def pytest_runtest_teardown(item: Item, nextitem: Item):
#     """Called after the test function has been called."""
# Stop the coverage collection
# Stop the trace
# sys.settrace(None)


#     internal_coverage.stop()
#     internal_coverage.save()
#     # Get the coverage data
#     cov_data = internal_coverage.get_data()
#     # Analyze the coverage data
#     for filename in cov_data.measured_files():
#         # Get the analysis for the file
#         analysis = internal_coverage._analyze(filename)
#         # Get the list of executed lines
#         executed_lines = analysis.executed
#         # Get the list of statements (lines that could have been executed)
#         statements = analysis.statements
#         # Get the list of missing lines (statements that were not executed)
#         missing_lines = analysis.missing
#         print(f"Test: {item.nodeid}")
#         print(f"File: {filename}")
#         print(f"Executed lines: {executed_lines}")
#         print(f"Statements: {statements}")
#         print(f"Missing lines: {missing_lines}")
#     # Clear the coverage data for the next test
#     internal_coverage.erase()

#     }}}



"""Offer advice through the use of the LLM-Based mentoring system."""

import random
import socket
from typing import List, Optional

import openai
import validators
from rich.console import Console
from rich.markdown import Markdown
from rich.panel import Panel

from . import enumerations
from .exceptions import get_litellm_traceback


def load_litellm() -> None:
    """Load the litellm module."""
    # note that the purpose of this function is
    # to allow the loading of the litellm module
    # to take place in a separate thread, thus
    # ensuring that the main interface is not blocked
    global litellm  # noqa: PLW0602
    global completion  # noqa: PLW0603
    from litellm import completion


def validate_url(value: str) -> bool:
    """Validate a URL given as a string using the validators library."""
    if not validators.url(value):
        return False
    return True


def handle_connection_error(console: Console) -> None:
    """Handle connection error."""
    # Print an error message stating there's issues with connecting to the api server.
    console.print(
        "[bold red]Error: Unable to connect to the API server.[/bold red]"
    )
    # Print a troubleshooting message.
    console.print(
        "Please check your network connection and ensure the API server is reachable."
    )


def check_internet_connection(timeout: int = 5) -> bool:
    """Check if the system has an active internet connection."""
    # List of well-known DNS servers to test connectivity
    dns_servers = [
        ("8.8.8.8", 53),  # Google DNS
        ("1.1.1.1", 53),  # Cloudflare DNS
        ("9.9.9.9", 53),  # Quad9 DNS
        ("208.67.222.222", 53),  # OpenDNS
    ]
    # Randomly select a DNS server from the list
    server = random.choice(dns_servers)
    try:
        # Attempt to create a socket connection to the selected DNS server
        socket.create_connection(server, timeout=timeout)
        # If the connection is successful, return True indicating internet is available.
        return True
    # If an OSError is raised, it indicates that the connection attempt failed.
    # This could be due to no internet connection or network issues.
    except OSError:
        # Return False indicating that the internet connection is not available.
        return False


def check_advice_model(
    console: Console,
    report: Optional[List[enumerations.ReportType]],
    advice_model: str,
) -> None:
    """Check if the advice request is valid because a model was specified."""
    if (
        report is not None
        and (
            enumerations.ReportType.testadvice in report
            or enumerations.ReportType.all in report
        )
        and advice_model is None
    ):
        console.print()
        console.print(
            "[red]The --advice-model option is required when --report includes 'advice' or 'all'"
        )


def check_advice_server(
    console: Console,
    report: Optional[List[enumerations.ReportType]],
    advice_method: str,
    advice_server: str,
) -> None:
    """Check if the advice request is valid because a server was specified."""
    if (
        report is not None
        and (
            enumerations.ReportType.testadvice in report
            or enumerations.ReportType.all in report
        )
        and advice_method == enumerations.AdviceMethod.api_server
        and advice_server is None
    ):
        console.print()
        console.print(
            "[red]The --advice-server option is required when --advice-method is 'api_server'"
        )
    elif (
        report is not None
        and (
            enumerations.ReportType.testadvice in report
            or enumerations.ReportType.all in report
        )
        and advice_method == enumerations.AdviceMethod.api_server
        and not validate_url(advice_server)
    ):
        console.print()
        console.print(
            "[bold red]Before sending to LLM:\nThe --advice-server option did not specify a valid URL"
        )


def fix_failures(  # noqa: PLR0913
    console: Console,
    filtered_test_output: str,
    exec_exam_test_assertion_details: str,
    test_overview: str,
    failing_test_details: str,
    failing_test_code: str,
    advice_method: enumerations.AdviceMethod,
    advice_model: str,
    advice_server: str,
    syntax_theme: enumerations.Theme,
    fancy: bool = True,
):
    """Offer advice through the use of the LLM-based mentoring system."""
    if not check_internet_connection():
        # If there is no internet connection, handle the connection error.
        # Call the handle_connection_error function
        handle_connection_error(console)
        return
    try:
        with console.status(
            "[bold green] Getting Feedback from ExecExam's Coding Mentor"
        ):
            test_overview = (
                filtered_test_output + exec_exam_test_assertion_details
            )
            llm_debugging_request = (
                "I am an undergraduate student completing a programming examination."
                + " You may never make suggestions to change the source code of the test cases."
                + " Always make suggestions about how to improve the Python source code of the program under test."
                + " Always give Python code in a Markdown fenced code block with your suggested program."
                + " Always start your response with a friendly greeting and overview of what you will provide."
                + " Always conclude by saying that you are making a helpful suggestion but could be wrong."
                + " Always be helpful, upbeat, friendly, encouraging, and concise when making a response."
                + " Your task is to suggest, in a step-by-step fashion, how to fix the bug(s) in the program?"
                + f" Here is the test overview with test output and details about test assertions: {test_overview}"
                + f" Here is a brief overview of the test failure information: {failing_test_details}"
                + f" Here is the source code for the one or more failing test(s): {failing_test_code}"
            )

            if advice_method == enumerations.AdviceMethod.api_key:
                # Submit the debugging request to the LLM-based mentoring system
                response = completion(  # type: ignore
                    model=advice_model,
                    messages=[
                        {"role": "user", "content": llm_debugging_request}
                    ],
                )
                # Display the advice from the LLM-based mentoring system
                if fancy:
                    console.print(
                        Panel(
                            Markdown(
                                str(response.choices[0].message.content),  # type: ignore
                            ),
                            expand=False,
                            title="Advice from ExecExam's Coding Mentor (API Key)",
                            padding=1,
                        )
                    )
                else:
                    console.print(
                        Markdown(
                            str(response.choices[0].message.content),  # type: ignore
                        ),
                    )
                    console.print()

            elif advice_method == enumerations.AdviceMethod.api_server:
                # Use the OpenAI approach to submit the debugging request
                client = openai.OpenAI(
                    api_key="anything", base_url=advice_server
                )
                response = client.chat.completions.create(
                    model=advice_model,
                    messages=[
                        {"role": "user", "content": llm_debugging_request}
                    ],
                )
                if fancy:
                    console.print(
                        Panel(
                            Markdown(
                                str(response.choices[0].message.content),
                                code_theme=syntax_theme.value,
                            ),
                            expand=False,
                            title="Advice from ExecExam's Coding Mentor (API Server)",
                            padding=1,
                        )
                    )
                else:
                    console.print(
                        Markdown(
                            str(response.choices[0].message.content),  # type: ignore
                        ),
                    )
                    console.print()
    except Exception:
        get_litellm_traceback(console)



"""Extract contents from data structures."""

from pathlib import Path
from typing import Any, Dict, List, Tuple

from . import convert


def is_failing_test_details_empty(details: str) -> bool:
    """Determine if the string contains a newline as a hallmark of no failing tests."""
    if details == "\n":
        return True
    return False


def extract_details(details: Dict[Any, Any]) -> str:
    """Extract the details of a dictionary and return it as a string."""
    output = []
    # iterate through the dictionary and add each key-value pair
    for key, value in details.items():
        output.append(f"{value} {key}")
    if len(output) == 0:
        return ""
    return "Details: " + ", ".join(output)


def extract_test_run_details(details: Dict[Any, Any]) -> str:
    """Extract the details of a test run."""
    # Format of the data in the dictionary:
    # 'summary': Counter({'passed': 2, 'total': 2, 'collected': 2})
    summary_details = details["summary"]
    # convert the dictionary of summary to a string
    summary_details_str = extract_details(summary_details)
    return summary_details_str


def extract_test_assertion_details(test_details: Dict[Any, Any]) -> str:
    """Extract the details of a dictionary and return it as a string."""
    # create an empty list to store the output
    output = []
    # indicate that this is the first assertion
    # to be processed (it will have a "-" to start)
    first = True
    # iterate through the dictionary and add each key-value pair
    # that contains the details about the assertion
    for key, value in test_details.items():
        # this is the first assertion and thus
        # the output will start with a "-"
        if first:
            output = ["  - "]
            output.append(f"{key}: {value}\n")
            first = False
        # this is not the first assertion and thus
        # the output will start with a "  " to indent
        else:
            output.append(f"    {key}: {value}\n")
    # return each index in the output list as a string
    return "".join(output)


def extract_test_assertion_details_list(details: List[Dict[Any, Any]]) -> str:
    """Extract the details of a list of dictionaries and return it as a string."""
    output = []
    # iterate through the list of dictionaries and add each dictionary
    # to the running string that conatins test assertion details
    for current_dict in details:
        output.append(extract_test_assertion_details(current_dict))
    return "".join(output)


def extract_test_assertions_details(test_reports: List[dict[str, Any]]):
    """Extract the details of test assertions."""
    # create an empty list that will store details about
    # each test case that was execued and each of
    # the assertions that was run for that test case
    test_report_string = ""
    # iterate through the list of test reports
    # where each report is a dictionary that includes
    # the name of the test and the assertions that it ran
    for test_report in test_reports:
        # get the name of the test
        test_name = test_report["nodeid"]
        # extract only the name of the test file and the test name,
        # basically all of the content after the final slash
        display_test_name = test_name.rsplit("/", 1)[-1]
        test_report_string += f"\n{display_test_name}\n"
        # there is data about the assertions for this
        # test and thus it should be extracted and reported
        if "assertions" in test_report:
            test_report_string += extract_test_assertion_details_list(
                test_report["assertions"]
            )
    # return the string that contains all of the test assertion details
    return test_report_string


def extract_failing_test_details(
    details: dict[Any, Any],
) -> Tuple[str, List[Dict[str, Path]]]:
    """Extract the details of a failing test."""
    # extract the tests from the details
    tests = details["tests"]
    # create an empty string that starts with a newline;
    # the goal of the for loop is to incrementally build
    # of a string that contains all deteails about failing tests
    failing_details_str = "\n"
    # create an initial path for the file containing the failing test
    failing_test_paths = []
    # incrementally build up results for all of the failing tests
    for test in tests:
        if test["outcome"] == "failed":
            current_test_failing_dict = {}
            # convert the dictionary of failing details to a string
            # and add it to the failing_details_str
            failing_details = test
            # get the nodeid of the failing test
            failing_test_nodeid = failing_details["nodeid"]
            failing_details_str += f"  Name: {failing_test_nodeid}\n"
            # get the call information of the failing test
            failing_test_call = failing_details["call"]
            # get the crash information of the failing test's call
            failing_test_crash = failing_test_call["crash"]
            # extract the root of the report, which corresponds
            # to the filesystem on which the tests were run
            failing_test_path_root = details["root"]
            # extract the name of the file that contains the test
            # from the name of the individual test case itself
            failing_test_nodeid_split = failing_test_nodeid.split("::")
            # create a complete path to the file that contains the failing test file
            failing_test_path = (
                Path(failing_test_path_root) / failing_test_nodeid_split[0]
            )
            # extract the name of the function from the nodeid
            failing_test_name = failing_test_nodeid_split[-1]
            # assign the details about the failing test to the dictionary
            current_test_failing_dict["test_name"] = failing_test_name
            current_test_failing_dict["test_path"] = failing_test_path
            failing_test_paths.append(current_test_failing_dict)
            # creation additional diagnotics about the failing test
            # for further display in the console in a text-based fashion
            failing_test_path_str = convert.path_to_string(
                failing_test_path, 4
            )
            failing_test_lineno = failing_test_crash["lineno"]
            failing_test_message = failing_test_crash["message"]
            # assemble all of the failing test details into the string
            failing_details_str += f"  Path: {failing_test_path_str}\n"
            failing_details_str += f"  Line number: {failing_test_lineno}\n"
            failing_details_str += f"  Message: {failing_test_message}\n"
    # return the string that contains all of the failing test details
    return (failing_details_str, failing_test_paths)


def extract_test_output(keep_line_label: str, output: str) -> str:
    """Filter the output of the test run to keep only the lines that contain the label."""
    # create an empty string that will store the filtered output
    filtered_output = ""
    # iterate through the lines in the output
    for line in output.splitlines():
        # if the line contains the label, add it to the filtered output
        if keep_line_label in line:
            filtered_output += line + "\n"
    # return the filtered output
    return filtered_output


def extract_test_output_multiple_labels(
    keep_line_labels: List[str], output: str
) -> str:
    """Filter the output of the test run to keep only the lines that contain the label."""
    # create an empty string that will store the filtered output
    filtered_output = ""
    # iterate through the lines in the output
    for line in output.splitlines():
        # if the line contains any one of the the labels,
        # then add it to the filtered output
        if any(label in line for label in keep_line_labels):
            filtered_output += line + "\n"
    # return the filtered output
    return filtered_output



"""Define exceptions for the input errors in the command line."""

import sys

# from rich.console import Console (ask if i need to import this?) i dont think i do because i pass it as an atg


def get_litellm_traceback(console) -> None:
    """Print the traceback of the last exception."""
    exc_type, exc_obj, _ = sys.exc_info()

    if exc_type is None:
        return
    # List of litellm exception types and their explanations
    litellm_exceptions = {
        "NotFoundError": "LLM resource not found. Please check your model and/or endpoint.",
        "AuthenticationError": "API authentication failed. Please verify your API key.",
        "RateLimitError": "Rate limit exceeded. Wait and retry or check API key.\nNOTE: This error can sometimes be caused by an invalid API key.",
        "InvalidRequestError": "Malformed API request. Please review parameters.",
        "APIError": "Internal LLM API error. Retry later.",
        "APIConnectionError": "Connection failed. \nNOTE: This error can sometimes be caused by an invalid server URL. Verify your server URL.",
    }

    # if statements to display exceptions
    if exc_type.__name__ in litellm_exceptions:
        console.print(
            f"[bold red]Exception Type: {exc_type.__name__}[/bold red]"
        )
        console.print(f"Explanation: {litellm_exceptions[exc_type.__name__]}")
    else:
        # default behavior for non-litellm exceptions
        console.print(
            f"[bold red]Exception Type: {exc_type.__name__}[/bold red]"
        )
        console.print(f"Error Message: {exc_obj!s}")

    # general purpose ouput as a backup
    console.print(
        "\n[bold red]If your issue persists, ensure the model you entered is correct, such as:[/bold red]"
    )
    console.print("[bold blue]- anthropic/claude-3-haiku-20240307[/bold blue]")
    console.print("[bold blue]- anthropic/claude-3-opus-20240229[/bold blue]")
    console.print("[bold blue]- groq/llama3-8b-8192[/bold blue]")
    console.print(
        "[bold blue]- openrouter/meta-llama/llama-3.1-8b-instruct:free[/bold blue]"
    )

    console.print(
        "\n[bold red]Please visit [bold blue]https://docs.litellm.ai/docs/providers [/bold blue]for more valid LiteLLM models[bold red]"
    )

    console.print(
        "\n[bold red]For server connectivity issues, please visit [bold blue]https://docs.litellm.ai/docs/simple_proxy [/bold blue]for a valid LiteLLM proxy.[/bold red]"
    )



"""Utility functions to record and produce debugging logs."""

from enum import Enum
from typing import List

# list of debugging messages
messages: List[str] = []


class Debug(str, Enum):
    """An enumeration of the various debugging messages."""

    get_advice_with_llm = (
        "[green]\u2714 Correctly requested and received advice from an LLM."
    )
    parameter_check_passed = (
        "[green]\u2714 Validity check passed for command-line arguments."
    )
    pytest_passed_with_marks = (
        "[green]\u2714 Correctly ran pytest when using marks."
    )
    pytest_passed_without_marks = (
        "[green]\u2714 Correctly ran pytest when not using marks."
    )
    started_litellm_thread = "[green]\u2714 Correctly started LiteLLM thread."
    stopped_litellm_thread = "[green]\u2714 Correctly stopped LiteLLM thread."
    started_capturing_output = (
        "[green]\u2714 Started to capture standard output and error."
    )
    stopped_capturing_output = (
        "[green]\u2714 Stopped capturing standard output and error."
    )


def debug(allow: bool, message: str) -> None:
    """Record a debugging message."""
    if allow:
        messages.append(message)


def has_debugging_messages() -> bool:
    """Determine if there are debugging messages."""
    return len(messages) > 0


def get_debugging_messages() -> str:
    """Retrieve a formatted version of the debugging messages."""
    # there are debugging messages; create a single
    # string with newlines at the start and the end
    # of the block of debugging messages
    if messages:
        all_messages = "\n" + "\n".join(messages)
        return all_messages + "\n"
    # there are no debugging messages and thus
    # this function must return an empty string
    return ""



"""Run an executable examination."""

import io
import os
import subprocess
import sys
import threading
import time
import warnings
from pathlib import Path
from typing import List, Optional

import pytest
import typer
from pytest_jsonreport.plugin import JSONReport  # type: ignore
from rich.console import Console
from typing_extensions import Annotated

from . import advise, display, enumerations, extract, util
from . import debug as debugger
from . import pytest_plugin as exec_exam_pytest_plugin

# suppress the warnings that are produced by the Pydantic library;
# note that this is needed because one of execexam's dependencies
# is not using Pydantic correctly and this produces warnings
warnings.filterwarnings("ignore", category=UserWarning, module="pydantic")

# create a Typer object to support the command-line interface
cli = typer.Typer(no_args_is_help=True)

# create a default console
console = Console()

# create the skip list for data not needed
skip = ["keywords", "setup", "teardown"]

# create a variable of the main pytest issues
pytest_labels = ["FAILED", "ERROR", "WARNING", "COLLECTERROR"]


def tldr_callback(value: bool) -> None:
    """Display a list of example commands and their descriptions."""
    if value:
        display.display_tldr(console)
        raise typer.Exit()


@cli.command()
def run(  # noqa: PLR0913, PLR0915
    project: Path = typer.Argument(
        ...,
        help="Project directory containing questions and tests",
    ),
    tests: Path = typer.Argument(
        ...,
        help="Test file or test directory",
    ),
    tldr: Annotated[
        Optional[bool],
        typer.Option(
            "--tldr",
            callback=tldr_callback,
            help="Display summary of commands",
        ),
    ] = None,
    report: Optional[List[enumerations.ReportType]] = typer.Option(
        None,
        help="Types of reports to generate",
    ),
    mark: str = typer.Option(None, help="Run tests with specified mark(s)"),
    maxfail: int = typer.Option(
        10, help="Maximum test failures before stopping"
    ),
    advice_method: enumerations.AdviceMethod = typer.Option(
        enumerations.AdviceMethod.api_key, help="LLM-based method for advice"
    ),
    advice_model: str = typer.Option(
        None, help="LLM model: https://docs.litellm.ai/docs/providers"
    ),
    advice_server: str = typer.Option(None, help="URL of the LiteLLM server"),
    debug: bool = typer.Option(False, help="Collect debugging information"),
    fancy: bool = typer.Option(True, help="Display fancy output"),
    syntax_theme: enumerations.Theme = typer.Option(
        enumerations.Theme.ansi_dark, help="Syntax highlighting theme"
    ),
) -> None:
    """Run an executable exam and produce the requested report(s)."""
    # indicate that the program's exit code is zero
    # to show that the program completed successfully;
    # attempt to prove otherwise by running all the checks
    return_code = 0
    # confirm that the advice model is provided when
    # the report includes the advice report type or
    # when the report includes all of the report types
    advise.check_advice_model(console, report, advice_model)
    # confirm that the advice server is provided when
    # the advice method is set to the API server and
    # the report includes the advice report type or all reports
    advise.check_advice_server(console, report, advice_method, advice_server)
    # load the litellm module in a separate thread when advice
    # was requested for this run of the program
    debugger.debug(debug, debugger.Debug.parameter_check_passed.value)
    litellm_thread = threading.Thread(target=advise.load_litellm)
    # if --tldr was specified, then display the TLDR summary
    # of the commands and then exit the program
    if tldr is not None:
        return
    # if execexam was configured to produce the report for advice
    # or if it was configured to produce all of the possible reports,
    # then start the litellm thread that provides the advice
    display_report_type = enumerations.ReportType.testadvice
    if report is not None and (
        display_report_type in report or enumerations.ReportType.all in report
    ):
        litellm_thread.start()
        debugger.debug(debug, debugger.Debug.started_litellm_thread.value)
    # add the project directory to the system path
    sys.path.append(str(project))
    # create the plugin that will collect all data
    # about the test runs and report it as a JSON object;
    # note that this approach avoids the need to write
    # a custom pytest plugin for the executable examination
    json_report_plugin = JSONReport()
    # display basic diagnostic information about command-line's arguments;
    # extract the local parmeters and then make a displayable string of them
    args = locals()
    colon_separated_diagnostics = display.make_colon_separated_string(args)
    # --> SETUP
    syntax = False
    newline = True
    display.display_content(
        console,
        enumerations.ReportType.setup,
        report,
        colon_separated_diagnostics,
        "Parameter Information",
        fancy,
        syntax,
        syntax_theme,
        "python",
        newline,
    )
    # run pytest for either:
    # - a single test file that was specified in tests
    # - a directory of test files that was specified in tests
    # note that this relies on pytest correctly discovering
    # all of the test files and running their test cases
    # redirect stdout and stderr to /dev/null
    captured_output = io.StringIO()
    sys.stdout = captured_output
    sys.stderr = captured_output
    debugger.debug(debug, debugger.Debug.started_capturing_output.value)
    # run pytest in a fashion that will not
    # produce any output to the console
    found_marks_str = mark
    # there were test marks on the command-line and
    # thus they should be run for the specified tests
    # (note that marks can control which tests are run)
    pytest_exit_code = 0
    if found_marks_str:
        pytest_exit_code = pytest.main(
            [
                "-q",
                "-ra",
                "-s",
                "-p",
                "no:logging",
                "-p",
                "no:warnings",
                "--tb=no",
                "--json-report-file=none",
                f"--maxfail={maxfail}",
                "-m",
                found_marks_str,
                os.path.join(tests),
            ],
            plugins=[json_report_plugin, exec_exam_pytest_plugin],
        )
        debugger.debug(debug, debugger.Debug.pytest_passed_with_marks.value)
    # there were no test marks specified on the command-line
    # and thus all of the tests should be run based on the specified
    # test file or test directory, which this provides to pytest
    else:
        pytest_exit_code = pytest.main(
            [
                "-q",
                "-ra",
                "-s",
                "-p",
                "no:logging",
                "-p",
                "no:warnings",
                "--tb=no",
                f"--maxfail={maxfail}",
                "--json-report-file=none",
                os.path.join(tests),
            ],
            plugins=[json_report_plugin, exec_exam_pytest_plugin],
        )
        debugger.debug(debug, debugger.Debug.pytest_passed_without_marks.value)
    # restore stdout and stderr; this will allow
    # the execexam program to continue to produce
    # output in the console
    sys.stdout = sys.__stdout__
    sys.stderr = sys.__stderr__
    debugger.debug(debug, debugger.Debug.stopped_capturing_output.value)
    # determine the return code for the execexam command
    # based on the exit code that was produced by pytest
    return_code = util.determine_execexam_return_code(pytest_exit_code)
    # extract the data that was created by the internal
    # execexam pytest plugin for further diagnostic display
    execexam_report = exec_exam_pytest_plugin.reports
    # extract the details about the test assertions
    # that come from the pytest plugin that execexam uses
    exec_exam_test_assertion_details = extract.extract_test_assertions_details(
        execexam_report
    )
    # --> display details about the test runs
    _ = extract.extract_test_run_details(json_report_plugin.report)  # type: ignore
    # filter the test output and decide if an
    # extra newline is or is not needed
    filtered_test_output = extract.extract_test_output_multiple_labels(
        pytest_labels,
        captured_output.getvalue(),
    )
    # add an extra newline to the filtered output
    # since there is a failing test case to display
    if filtered_test_output != "":
        filtered_test_output = "\n" + filtered_test_output
    # indicate that the material that will be displayed
    # is not source code and thus does not need syntax highlighting
    # --> TRACE
    syntax = False
    newline = True
    display.display_content(
        console,
        enumerations.ReportType.testtrace,
        report,
        filtered_test_output + exec_exam_test_assertion_details,
        "Test Trace",
        fancy,
        syntax,
        syntax_theme,
        "python",
        newline,
    )
    # display details about the failing tests,
    # if they exist. Note that there can be:
    # - zero failing tests
    # - one failing test
    # - multiple failing tests
    # note that details about the failing tests are
    # collected by the execexam pytest plugin and
    # there is no need for the developer of the
    # examination to collect and report this data
    (
        failing_test_details,
        failing_test_path_dicts,
    ) = extract.extract_failing_test_details(json_report_plugin.report)  # type: ignore
    failing_test_code_overall = ""
    # there was at least one failing test case
    if not extract.is_failing_test_details_empty(failing_test_details):
        # display additional helpful information about the failing
        # test cases; this is the error message that would appear
        # when standardly running the test suite with pytest
        # --> FAILURE
        syntax = False
        newline = True
        display.display_content(
            console,
            enumerations.ReportType.testfailures,
            report,
            failing_test_details,
            "Test Failure(s)",
            fancy,
            syntax,
            syntax_theme,
            "Python",
            newline,
        )
        # display the source code for the failing test cases
        for failing_test_path_dict in failing_test_path_dicts:
            test_name = failing_test_path_dict["test_name"]
            failing_test_path = failing_test_path_dict["test_path"]
            # build the command for running symbex; this tool can
            # perform static analysis of Python source code and
            # extract the code of a function inside of a file
            command = f"symbex {test_name} -f {failing_test_path}"
            # run the symbex command and collect its output
            process = subprocess.run(
                command,
                shell=True,
                check=True,
                text=True,
                capture_output=True,
            )
            # delete an extra blank line from the end of the file
            # if there are two blank lines in a row
            sanitized_output = process.stdout.rstrip() + "\n"
            failing_test_code_overall += sanitized_output
            # display the source code of the failing test
            # --> CODE
            syntax = True
            newline = True
            display.display_content(
                console,
                enumerations.ReportType.testcodes,
                report,
                sanitized_output,
                "Failing Test",
                fancy,
                syntax,
                syntax_theme,
                "Python",
                newline,
            )
    # display the spinner until the litellm thread finishes
    # loading the litellm module that provides the LLM-based
    # mentoring by automatically suggesting fixes for test failures
    display_report_type = enumerations.ReportType.testadvice
    if report is not None and (
        display_report_type in report or enumerations.ReportType.all in report
    ):
        # regardless of whether or not there were test failures, it is
        # appropriate to display the loading message so that the thread
        # can finish the imports and then the advice can be requested
        # (if there were test failures) or the program can stop running
        # (if there were no test failures and thus advice is not needed)
        console.print()
        with console.status("[bold green] Loading ExecExam's Coding Mentor"):
            while litellm_thread.is_alive():
                time.sleep(0.1)
        # return control to the main thread now that the
        # litellm module has been loaded in a separate thread
        litellm_thread.join()
        debugger.debug(debug, debugger.Debug.stopped_litellm_thread.value)
        # provide advice about how to fix the failing tests
        # because the non-zero return code indicates that
        # there was a test failure and that overall there
        # is at least one mistake in the examination for
        # which advice should be sought from the LLM
        if return_code != 0:
            advise.fix_failures(
                console,
                filtered_test_output,
                exec_exam_test_assertion_details,
                filtered_test_output + exec_exam_test_assertion_details,
                failing_test_details,
                failing_test_code_overall,
                advice_method,
                advice_model,
                advice_server,
                syntax_theme,
                fancy,
            )
            debugger.debug(debug, debugger.Debug.get_advice_with_llm.value)
        # there were no test failures and thus there is no need
        # to seek advice from the LLM-based mentoring system;
        # display a message to repor that even though advice
        # was requested, it was not needed and thus is not displayed
        else:
            syntax = False
            newline = False
            advice_message = display.display_advice(return_code)
            display.display_content(
                console,
                enumerations.ReportType.exitcode,
                report,
                advice_message,
                "Advice Status",
                fancy,
                syntax,
                syntax_theme,
                "Python",
                newline,
            )
    # display the debugging messages
    debugging_messages_exist = debugger.has_debugging_messages()
    if debugging_messages_exist:
        debugging_messages = debugger.get_debugging_messages()
        syntax = False
        newline = True
        display.display_content(
            console,
            enumerations.ReportType.debug,
            report,
            debugging_messages,
            "Debugging Information",
            fancy,
            syntax,
            syntax_theme,
            "Python",
            newline,
        )
    # display a final message about the return code, using
    # a human-readable message that indicates the overall status
    exit_code_message = display.get_display_return_code(return_code, fancy)
    # display the return code through a diagnostic message
    syntax = False
    newline = True
    display.display_content(
        console,
        enumerations.ReportType.exitcode,
        report,
        exit_code_message,
        "Overall Status",
        fancy,
        syntax,
        syntax_theme,
        "Python",
        newline,
    )
    # return the code for the overall success of the program
    # to communicate to the operating system the examination's status
    sys.exit(return_code)



"""Utility functions for the execexam package."""

import pytest


def determine_execexam_return_code(pytest_return_code: int) -> int:
    """Determine the return code for the execexam command by pytest code."""
    execexam_return_code = 0
    # see if the pytest exit code is one of the designated
    # codes and then assign it to the appropriate value
    if pytest_return_code == pytest.ExitCode.TESTS_FAILED:
        execexam_return_code = 1
    elif pytest_return_code == pytest.ExitCode.INTERRUPTED:
        execexam_return_code = 2
    elif pytest_return_code == pytest.ExitCode.INTERNAL_ERROR:
        execexam_return_code = 3
    elif pytest_return_code == pytest.ExitCode.USAGE_ERROR:
        execexam_return_code = 4
    elif pytest_return_code == pytest.ExitCode.NO_TESTS_COLLECTED:
        execexam_return_code = 5
    return execexam_return_code
